
#include <stdio.h>

#define BLK1 0xff
#define BLK2 0xffff
#define BLK3 0xffffff
#define BLK4 0xffffffff

unsigned int filesize(char *name,unsigned int *freq,int * max_run_len) {
    FILE * f;
    unsigned int s = 0;
    int i;
    unsigned char ch;
    unsigned int run_len[256];
    unsigned int cur_len = 0;
    unsigned char last_ch = 0;
    *max_run_len = 0;
    f = fopen(name,"rb");
    if (f == NULL)
	return 0;
    for (i=0;i<256;i++) {
	freq[i] = 0;	
	run_len[i] = 0;
    }	
    while (fread(&ch,sizeof(ch),1,f)!=0) {
	// printf("%d: %d(%x)\n",s,ch,ch);
	if (s == 0)
	    last_ch = ch;
	if (ch!=last_ch) {
	    if (run_len[last_ch]<cur_len) {
		run_len[last_ch] = cur_len;
		if (cur_len > *max_run_len)
		    *max_run_len = cur_len;
	    }
	    cur_len = 0;
	    last_ch = ch;
	}
	cur_len ++;
	freq[ch] ++;
	s ++;
    }	
    fclose(f);
    return s;
}

char find_markers(unsigned int *freq,unsigned char *markers,unsigned int max_run_len) {
    char tot_markers = 0;
    unsigned char cur_marker;
    unsigned int cur_marker_freq;
    unsigned char j;
    int i,k;
    char ok;
    int m[4];
    m[0] = m[1] = m[2] = m[3] = -1;
    for (j=0;j<4;j++) {
	cur_marker_freq = 0xffffffff;
	cur_marker = 0;
	for (i=0;i<256;i++) {
	    if (freq[i]<=cur_marker_freq) {
		// printf("%d) %d;%d => %d;%d\n",j,cur_marker,cur_marker_freq,i,freq[i]);
		if ((int)i != m[0] && (int)i!=m[1] && (int)i!=m[2] && (int)i!=m[3]) {
		    cur_marker_freq = freq[i];
		    cur_marker = i;
		}
	    }
	}
	markers[j] = cur_marker;	
	m[j] = markers[j];
	// printf("Marker %d = %d; freq = %d\n",j,markers[j],cur_marker_freq);
    }
    if (max_run_len<=BLK1)
	tot_markers = 1;
    else
    if (max_run_len<=BLK2)
	tot_markers = 2;
    else    
    if (max_run_len<=BLK3)
	tot_markers = 3;
    else
    if (max_run_len<=BLK4)
	tot_markers = 4;
    return tot_markers;
}

int write_block(FILE *f,unsigned char ch,unsigned int times,unsigned char *markers) {
    int i;
    char val;
    //printf("Writing %d(%x)[%d]\n",ch,ch,times);
    if (times < 3 && ch!=markers[0] && ch!=markers[1] && ch!=markers[2] && ch!=markers[3]) {
	for (i=0;i<times;i++)
	    fwrite(&ch,sizeof(ch),1,f);
	return 0;    
    }
    if (times <= BLK1) {
	fwrite(&markers[0],sizeof(markers[0]),1,f);
	fwrite(&ch,sizeof(ch),1,f);
	val = (char)(times & 0xff);
	fwrite(&val,sizeof(val),1,f);
	return 1;
    }
    if (times <= BLK2) {
	fwrite(&markers[1],sizeof(markers[1]),1,f);
	fwrite(&ch,sizeof(ch),1,f);
	val = (char)(times & 0xff);
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff00) >> 8;
	fwrite(&val,sizeof(val),1,f);
	return 2;    
    }
    if (times <= BLK3) {
	fwrite(&markers[2],sizeof(markers[2]),1,f);
	fwrite(&ch,sizeof(ch),1,f);
	val = (char)(times & 0xff);
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff00) >> 8;
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff0000) >> 16;
	fwrite(&val,sizeof(val),1,f);
	return 3;    
    }
    if (times <= BLK4) {
	fwrite(&markers[3],sizeof(markers[3]),1,f);
	fwrite(&ch,sizeof(ch),1,f);
	val = (char)(times & 0xff);
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff00) >> 8;
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff0000) >> 16;
	fwrite(&val,sizeof(val),1,f);
	val = (times & 0xff000000) >> 24;
	fwrite(&val,sizeof(val),1,f);
	return 4;    
    }
    return -1;
}

unsigned int compress(char *fname,char *outfile) {
    unsigned int fsize;
    unsigned int fsize_compr = 0;
    unsigned int freq[256];
    unsigned int max_run_len;
    unsigned char markers[4];
    char num_markers = 0;
    int i;
    FILE * f;
    FILE * f1;
    fsize = filesize(fname,freq,&max_run_len);
    unsigned char last_ch, ch;
    unsigned int runlen = 0;
    if (fsize==0)
	return 0;
    f = fopen(fname,"rb");
    if (f == NULL)
	return 0;
    f1= fopen(outfile,"wb");
    if (f1 == NULL) {
	fclose(f);
	return 0;
    }
    num_markers = find_markers(freq,markers,max_run_len);
    if (num_markers > 0) {
	fwrite(&fsize,sizeof(fsize),1,f1);
	fwrite(&num_markers,sizeof(num_markers),1,f1);
	for (i=0;i<num_markers;i++)
	    fwrite(&markers[i],sizeof(markers[i]),1,f1);
	fread(&ch,sizeof(ch),1,f);
	last_ch = ch;
	runlen = 1;
	while (fread(&ch,sizeof(ch),1,f)!=0) {
	    //printf("%x\n",ch);
	    if (ch != last_ch || runlen == BLK4) {
		write_block(f1,last_ch,runlen,markers);
		last_ch = ch;
		runlen = 0;
	    }
	    runlen ++;	    
	}
	if (runlen>0)
    		write_block(f1,last_ch,runlen,markers);
    }
    fclose(f);
    fclose(f1);
    return fsize_compr;
}

unsigned int decompress(char *fname,char *fout) {
    unsigned int filesize;
    unsigned char num_markers;
    unsigned char markers[4];
    unsigned char ch,val;
    unsigned int len;
    unsigned int i;
    FILE *f;
    FILE *f1;
    f = fopen(fname,"rb");
    if (f == NULL)
	return 0;
    f1 = fopen(fout,"wb");
    if (f1 == NULL) {
	fclose(f);
	return 0;
    }
    fread(&filesize,sizeof(filesize),1,f);
    fread(&num_markers,sizeof(num_markers),1,f);
    for (i=0;i<num_markers;i++)
	fread(&markers[i],sizeof(markers[i]),1,f);
    while (fread(&ch,sizeof(ch),1,f)!=0) {
	    if (ch == markers[0]) {
		fread(&ch,sizeof(ch),1,f);
		fread(&val,sizeof(val),1,f);
		len = val;
	    } else
	    if (ch == markers[1] && num_markers >= 2) {
		fread(&ch,sizeof(ch),1,f);
		fread(&val,sizeof(val),1,f);
		len = val;
		fread(&val,sizeof(val),1,f);
		len = len + (val << 8);
	    } else
	    if (ch == markers[2] && num_markers >= 3) {
		fread(&ch,sizeof(ch),1,f);
		fread(&val,sizeof(val),1,f);
		len = val;
		fread(&val,sizeof(val),1,f);
		len = len + (val << 8);	
		fread(&val,sizeof(val),1,f);
		len = len + (val << 16);	
	    } else
	    if (ch == markers[3] && num_markers == 4 ) {
		fread(&ch,sizeof(ch),1,f);
		fread(&val,sizeof(val),1,f);
		len = val;
		fread(&val,sizeof(val),1,f);
		len = len + (val << 8);	
		fread(&val,sizeof(val),1,f);
		len = len + (val << 16);			
		fread(&val,sizeof(val),1,f);
		len = len + (val << 24);			
	    }
	    else {
		len = 1;
	    }
	// printf("Writing %d x %d\n",ch,len);
	for (i=0;i<len;i++)
	    fwrite(&ch,sizeof(ch),1,f1);	
    }    
    fclose(f);
    fclose(f1);
    return filesize;
}


int main(int argc,char **argv) {

    if (strcmp(argv[1],"d")==0)
	printf("%d",decompress(argv[2],argv[3]));

    if (strcmp(argv[1],"c")==0)
        printf("%d",compress(argv[2],argv[3]));

    return 0;
}
